@@ -570,7 +570,9 @@ int replaceOperands(StringRef dst_desc, StringRef src_desc, Module &M)
 
     Value *SV;
     if (src_desc[0] == 'U' || src_desc[0] == 'A') {
-        SV = cast<Value>(walkExact(src_desc.str(), dummy, M, DV->getType(), false));
+        SV = dyn_cast_or_null<Value>(walkExact(src_desc.str(), dummy, M, DV->getType(), false));
+        if (SV == nullptr)
+            return -5;
         if (SV->getType()->isVoidTy())
             return -3;
         if (DV->getType() != SV->getType())
